/*

					Disassembler for Games, Mark 2

*/

#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <ctype.h>
#include "opcodes.h"

unsigned char ROM[4096];				/* Space for ROM */
unsigned char Type[4096];				/* What is there ? */

#define	ISDATA	(0x01)					/* True if data space */
#define ISCODE	(0x02)					/* True if code space */
#define ISLABEL	(0x03)					/* True if referenced by another */

int  Identify(int Address)
{
unsigned int Word,n;
n = 0;
while (1)								/* The last one MUST match ! */
	{
	Word = ROM[Address];				/* Get the instruction */
	if (Opcodes[n].Bytes == 2)
		Word = (Word << 8) | ROM[Address+1];
										/* Return if match found */
	if ((Word & Opcodes[n].Mask) ==
					Opcodes[n].Compare) return(n);
	n++;
	}
return(0);
}

void Trace(int Address)
{
int Short,Complete,OType,Byte,CByte,Addr,Imm;
OType = Type[Address];					/* Get old address type */
Type[Address] = ISLABEL;				/* It is now a label */
if (OType != ISDATA) return;			/* Return if it wasn't a new one */
Complete = 0;
while (Complete == 0)					/* if not completed then */
	{
	if (Type[Address] == ISDATA)		/* This location is code not data */
				Type[Address] = ISCODE;
	Byte = ROM[Address];				/* See what is there */
	CByte = (Byte & 0xF0);				/* Get just the instruction bit */

	if (CByte == 0x10) Complete = 1;	/* Completed if JMP found */
	if (Byte == 0xC0) Complete = 1;		/* Completed if RET found */

	Imm = ROM[Address+1];				/* Some useful things */
	Addr = Imm + ((Byte & 0x0F) << 8);
	Short = Imm + (Address & 0xFF00);

	if (CByte == 0x10) Trace(Addr);		/* Handle JMP */
	if (CByte == 0x20) Trace(Addr);		/* Handle CALL */
	if (CByte == 0x30) Trace(Short);	/* Handle JNZ */
	if (CByte == 0x40) Trace(Short);	/* Handle JZ */
	if (CByte == 0x50) Trace(Address+4);/* Handle SKNE */
	if (Byte == 0x70)  Trace(Short);	/* Handle DJNZ */
	if (CByte == 0xD0) Trace(Short);	/* Handle KEY */
	if (Byte == 0xE8)  Trace(Short);	/* Handle Draw Sprite */

	OType = Identify(Address);			/* Identify the instruction */

	if (Opcodes[OType].Bytes == 2)		/* If 2 bytes, mark 2nd Code */
		if (Type[Address+1] == ISDATA)
					Type[Address+1] = ISCODE;
	Address = Address +					/* Go to next instruction */
					Opcodes[OType].Bytes;
	}
}

void GetLabel(char *Buffer,int Addr)
{
sprintf(Buffer,"L_%03X",Addr);
}

void Expand(char *Buffer,char *Src,unsigned int Word,int Addr)
{
while (*Src != '\0')
	{
	if (*Src == '.')
		{
		Src++;
		switch(*Src++)
			{
			case 'x':	sprintf(Buffer,"%1x",(Word >> 8) & 15);break;
			case 'y':	sprintf(Buffer,"%1x",(Word >> 4) & 15);break;
			case 'i':	sprintf(Buffer,"$%02x",Word & 0xFF);break;
			case 'z':	sprintf(Buffer,"$%03x",Word & 0xFFF);break;
			case 's':	GetLabel(Buffer,(Addr & 0x0F00) | (Word & 0xFF));break;
			case 'a':	GetLabel(Buffer,Word & 0xFFF);break;
			default:	printf("**** %s\n",Src-2);exit(0);
			}
		Buffer = Buffer+strlen(Buffer);
		}
	else
		{
		*Buffer++ = tolower(*Src);Src++;
		}
	}
*Buffer = '\0';
}

void Decode(int *pAddr)
{
char Label[12],Line[40],Comment[40];
int  i,OType,Bytes;
unsigned int Word;

Label[0] = Line[0] = Comment[0] = '\0';	/* Clear strings */
if (Type[*pAddr] == ISLABEL)			/* if label, mark as such */
	{
	GetLabel(Label,*pAddr);
	strcat(Label,":");
	}

Bytes = 1;								/* Default for data */
if (Type[*pAddr] != ISDATA)				/* If code... */
	{
	OType = Identify(*pAddr);			/* Get instruction type */
	Bytes = Opcodes[OType].Bytes;		/* Get the number of bytes */
	Word = ROM[*pAddr];					/* Get instruction word */
	if (Bytes == 2)
		Word = (Word << 8) | ROM[*pAddr+1];
	Expand(Line,Opcodes[OType].Text,	/* Convert to text */
							Word,*pAddr);
	}
else
	{
	sprintf(Line,"byte $%02x",ROM[*pAddr]);
	}
sprintf(Comment,"; %04x %02x",*pAddr,ROM[*pAddr]);
if (Bytes == 2)
	sprintf(Comment+strlen(Comment),"%02x",ROM[*pAddr+1]);

printf("%-8s%-24s%s\n",Label,Line,Comment);
*pAddr = (*pAddr)+Bytes;
}

void main()
{
FILE *f;
int i;

memset(ROM,0xFF,sizeof(ROM));			/* Fill ROM with 0xFF */
memset(Type,ISDATA,sizeof(Type));		/* Fill 'Type' Buffer with ISDATA */
f = fopen("STUDIO2.ROM","rb");			/* Read in the Studio 2 ROM */
fread(ROM,1,sizeof(ROM),f);				/* Read in the bytes */
fclose(f);

Trace(0x300);							/* Trace code at the entry point */
Trace(0x336);
Trace(0x33D);
Trace(0x344);
Trace(0x35B);
Trace(0x3B0);
Trace(0x420);
Trace(0x426);
Trace(0x449);
Trace(0x44D);
Trace(0x4D3);
Trace(0x4E7);
Trace(0x4F5);
Trace(0x569);
Trace(0x5B9);
Trace(0x611);
Trace(0x679);
Trace(0x72A);
Trace(0x3D8);

i = 0x300;
while (i < 0x800)
	{
	Decode(&i);
	}
}
